/*
 *  InitPrevNodeList.c
 */

/*
 *  Initialize the list of previous nodes at each level.
 *  Three list-maintenance procedures GetPrevNodeAtLevel, SetPrevNodeAtLevel,
 *    and InitPrevNodeList, maintain a singly linked list.
 *  Each entry in the list corresponds to the node previous to the current
 *    node at a given level (for example, element 2 in the list corresponds
 *    to the node to the left of the current node at level 2).
 *  If the maximum tree size is known beforehand, this list can be replaced
 *    by a fixed-sized array, and these procedures become trivial.
 *  Each list element contains two fields:  PREVNODE - the previous node at
 *    this level - and NEXTLEVEL - a forward pointer to the next list element.
 *  The list does not need to be cleaned up between calls to PositionTree
 *    for performance.
 */

#include "PositionTree.h"

InitPrevNodeList()
{
    struct pn * TempPtr;

    /*  Start with the node at level 0 -- the apex of the tree.		*/
    TempPtr = LevelZeroPtr;
    while ( TempPtr != (struct pn *)0 )
	{
	PREVNODE( TempPtr ) = NULL;
	TempPtr = NEXTLEVEL( TempPtr );
	}
    return;
}


struct node *
GetPrevNodeAtLevel( Level )
int	Level;
{
    struct pn * TempPtr;
    int		i;

    /*  Start with the node at level 0 -- the apex of the tree.		*/
    TempPtr = LevelZeroPtr;
    i = 0;
    while ( TempPtr != (struct pn *)0 )
	{
	if ( i == Level )
	    return( PREVNODE( TempPtr ) );

	TempPtr = NEXTLEVEL( TempPtr );
	i++;
	}
    /*  Otherwise, there was no node at the specific level.		*/
    return( NULL );
}


/*  Set an element in the list.						*/

SetPrevNodeAtLevel( Level, Node )
int	Level;
struct node * Node;
{
    struct pn * NewNode;
    struct pn * TempPtr;
    int		i;

    /*  Start with the node at level 0 -- the apex of the tree.		*/
    TempPtr = LevelZeroPtr;
    i = 0;
    while ( TempPtr != (struct pn *)0 )
	{
	if ( i == Level )
	    {
	    /*  At this level, replace the existing list		*/
	    /*    element with the passed-in node.			*/
	    PREVNODE( TempPtr ) = Node;
	    return;
	    }
	else if ( NEXTLEVEL( TempPtr ) == (struct pn *)0 )
	    /*  There isn't a list element yet at this level, so	*/
	    /*    add one.  The following instructions prepare the	*/
	    /*    list element at the next level, not at this one.	*/
	    {
	    NewNode = (struct pn *)malloc( sizeof( struct pn ) );
	    PREVNODE( NewNode ) = NULL;
	    NEXTLEVEL( NewNode ) = (struct pn *)0;
	    NEXTLEVEL( TempPtr ) = NewNode;
	    }
	/*  Prepare to move to the next level, to look again.		*/
	TempPtr = NEXTLEVEL( TempPtr );
	i++;
	}

    /*  Should only get here if LevelZeroPtr is nil.			*/
    LevelZeroPtr = (struct pn *)malloc( sizeof( struct pn ) );
    PREVNODE( LevelZeroPtr ) = Node;
    NEXTLEVEL( LevelZeroPtr ) = (struct pn *)0;
    return;
}
