/*
 *  xjsv.c	DCM Wood  93/03/05
 *
 *  Read networks (and subnets).  Store them with hashed pointers to them.
 *  Read interfaces.  Store them with hashed pointer to start of each subnet.
 *  For each interface,
 *    Increment the count of interfaces for the appropriate net and subnet
 *  For each subnet,
 *    Increment the count of subnets for the appropriate net
 *
 *  Make Nets Display
 *    For each net,
 *      If interface count > 0 or subnet sount > 1,
 *        MakeButton with action to display subnets
 *	      (May need an immediate message "Building subnet display")
 *	else
 *	  MakeButton with action to display
 *	      "No further info available for this network"  and beep
 *
 *  Make Subnets Display
 *    For each subnet of selected network,
 *      If interface count > 0,
 *        MakeButton with action to Make Interface Display
 *	      (May need an immediate message "Building interface display")
 *	else
 *	  MakeButton with action to display
 *	      "No further info available for this subnet"  and beep
 *
 *  Make Interface Display
 *    For each interface on selected subnet,
 *      Create Display for that interface showing
 *	  IP addr, Name, Ether Addr, Subnet Mask
 *    MakeButton with action to return to Subnet display
 *
 *
 *  Would like to be able to click on an interace that is part of a gateway
 *    The display would then show all the interfaces/subnets for that gateway
 *    The user could then click on any of the interfaces/subnets to look at
 *      the interface list for that subnet.
 */

#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <sys/types.h>
#include <netinet/in.h>

#include "js.h"

#define MAIN
#include "xjsv.h"

#include "jsdb.h"

static void MainLoop();
static void PutUp();
static void SetSubNets();
static void TakeDown();
static void Usage();

QuitStat();

static void MakeNetBtns();
static void ShowNetBtns();
static void HideNetBtns();

static struct Btn * ButQuit;

static int NewState = NETS_STATE;

static int WindowWidth = 400;
static int WindowHeight = 100;

struct Node *	NewNet = NULL;		/* Network selected */
struct Node *	NewSubnet = NULL;	/* Subnet selected */
struct Node *	NewIntf = NULL;		/* Interface selected */

struct Node *	CurNet = NULL;		/* Network selected */
struct Node *	CurSubnet = NULL;	/* Subnet selected */
struct Node *	CurIntf = NULL;		/* Interface selected */

struct Btn *	BHT[HASH_TABLE_SIZE];
struct Node *	NHT[HASH_TABLE_SIZE];
struct Node *	IHT[HASH_TABLE_SIZE];

struct Btn * BtnOne = NULL;		/* Linked list of buttons */
struct Btn * BtnLast = NULL;

struct Node * fp[MAX_NTYPE];		/* Linked list of nodes by ntype */
struct Node * lp[MAX_NTYPE];

int	MainLoopState = NETS_STATE;
int	debug = 0;			/* debug flag */


static void
Catch()
{
    fprintf( stderr, " QUIT signal caught - terminating xjsv\n" );
    exit( 0 );
}


main( argc, argv )
int	argc;
char **	argv;
{
    char *  dsnp;
    int     id;
    int     i;

    signal( SIGQUIT, Catch );
    setlinebuf( stdout );
    gethostname( hostname, MAXHOSTNAMELEN );
    jshostp = getenv( "JS_HOST" );

#ifdef DEMO
    if ( jshostp == NULL )
	jshostp = "bruno.cs.colorado.edu.";
#endif 

    /* get the command line arguments if they exist */

    while ( *(++argv) )
	{
	if ( **argv != '-' )
	    {
	    printf( "Use a - before arguments\n" );
	    exit( 1 );
	    }
	switch ( *(*argv + 1) )
	    {
	    case 'd':
		debug = 1;	/* set the debug flag */
		break;

	    case 'j':
		if ( *(++argv) )
		    jshostp = argv[0];
		else
		    {
		    Usage();
		    exit( -1 );
		    }
		break;

	    default:
		Usage();
		exit( -1 );
	    }
	}

    if ( jshostp == NULL )
	jshostp = hostname;
    fprintf( stderr, " Using Journal Server at %s\n", jshostp );
    ReadJS( INTF );
    ReadJS( NETWORK );
    ReadJS( GATEWAY );
    FixIntfCounts();

    dsnp = getenv( "DISPLAY" );
    fprintf( stderr, " Using  %s  for X-display\n", dsnp );

    InitWin( hostname );

    MakeNetBtns();
    ResizeWindow( WindowWidth, WindowHeight );
    ShowNetBtns();

    /* SetHost( jshostp ); */

    FlushDpy();

    if ( debug )
	printf( " Entering MainLoop\n" );
    MainLoop();

}


static void
MainLoop()
{
    while( 1 )
	{

	if( NewState != MainLoopState )
	    {
	    TakeDown( MainLoopState );
	    MainLoopState = NewState;
	    PutUp( NewState );
	    }

	DoEvents( TRUE );
	}
}


void
ReturnToMain()
{
    switch ( MainLoopState )
	{
	case NETS_STATE:
	    exit( 0 );
	    break;

	case SUBS_STATE:
	    NewState = NETS_STATE;
	    break;

	case INTF_STATE:
	    NewState = SUBS_STATE;
	    break;

	case INDV_STATE:
	    NewState = INTF_STATE;
	    break;
	}
}


static void
TakeDown( state )
int state;
{
    switch( state )
	{
	case NETS_STATE:
	    HideNetBtns();
	    break;

	case SUBS_STATE:
	    HideSubsWindow();	
	    break;

	case INTF_STATE:
	    HideIntfWindow();
	    break;

	case INDV_STATE:
	    HideIndvWindow();
	    break;
	}
}


static void
PutUp( state )
int state;
{
    switch( state )
	{
	case NETS_STATE:
	    ResizeWindow( WindowWidth, WindowHeight );
	    ShowNetBtns();
	    break;

	case SUBS_STATE:
	    if ( NewNet != CurNet )
		{
		if ( CurNet != NULL )
		    WipeSubsDisplay();
		CurNet = NewNet;
		MakeSubsDisplay();
		}
	    ShowSubsWindow();	
	    break;

	case INTF_STATE:
	    if ( NewSubnet != CurSubnet )
		{
		if ( CurSubnet != NULL )
		    WipeIntfDisplay();
		CurSubnet = NewSubnet;
		MakeIntfDisplay();
		}
	    ShowIntfWindow();
	    break;

	case INDV_STATE:
	    if ( NewIntf != CurIntf )
		{
		if ( CurIntf != NULL )
		    WipeIndvDisplay();
		CurIntf = NewIntf;
		MakeIndvDisplay();
		}
	    ShowIndvWindow();
	    break;
	}
}


static void
MakeNetBtns()
{
    static int ncol = 1;
    static int xpos = 0;
    static int ypos = 50;
    static int count = 0;

    struct Btn * bp;
    char    buf[256];
    int     i;
    int     len;
    struct Node * np;

    if ( debug )
	printf( " MakeNetBtns\n" );

    ButQuit = CreateBtn( "Quit", 0, 15, 5, 5, QuitStat, NULL );
    ShowBtn( ButQuit );

    GetNets();
    for ( i=0; i<MAX_BTNS; i++ )
	{
	np = GetNextNet();
	if ( np == NULL )
	    break;

	bp = CreateBtn( NULL, 200, 17, xpos, ypos, SetSubNets, np );
	if( count < 50 )
	    {
	    ypos += 15;
	    count++;
	    }
	else 
	    {
	    ncol++;
	    xpos += 200;
	    ypos = 50;
	    count = 0;
	    }
	if ( ypos > WindowHeight )
	    WindowHeight = ypos + 5;
	}
    FlushDpy();
}


static void
HideNetBtns()
{
    struct Btn * bp;

    /* HideBtn( ButQuit ); */

    for ( bp=BtnOne; bp!=NULL; bp=bp->next )
	HideBtn( bp );

    FlushDpy();
}


static void
ShowNetBtns()
{
    struct Btn * bp;
    char    cbuf[100];

    ShowBtn( ButQuit );

    sprintf( cbuf, "Networks known by Journal Server at %s", jshostp );
    DrawTitle( cbuf );

    for ( bp=BtnOne; bp!=NULL; bp=bp->next )
	{
	ShowBtn( bp );
	/* RedrawBtn( bp ); */
	}

    FlushDpy();
}


QuitStat()
{
    exit( 0 );
}


static void
SetSubNets( bp )
struct Btn * bp;
{
    if ( debug )
	printf( " SetSubNets( bp = %X)\n" );

    NewNet = bp->np;
    NewState = SUBS_STATE;
}


void
SetIntf( bp )
struct Btn * bp;
{
    if ( debug )
	printf( " SetIntf( bp = %X)\n" );

    NewSubnet = bp->np;
    NewState = INTF_STATE;
}


void
SetIndv( bp )
struct Btn * bp;
{
    if ( debug )
	printf( " SetIndv( bp = %X)\n" );

    NewIntf = bp->np;
    NewState = INDV_STATE;
}


void
SetSN( bp )
struct Btn * bp;
{
    u_long  nm;
    struct Node * np;
    struct Node * netp;
    struct Node * snp;

    if ( debug )
	printf( " SetSN( bp = %X)\n" );

    np = bp->np;
    if ( np == NULL )
	return;

    nm = GetNetMask( np->addr );
    netp = FindNodeEnt( np->addr & nm, NTYPE_NET );
    if ( netp == NULL )
	return;
    
    if ( netp->sm != 0 )
	snp = FindNodeEnt( np->addr & netp->sm, NTYPE_SUBNET );

    if ( snp == NULL )
	return;

    NewSubnet = snp;
    NewState = INTF_STATE;
    return;
}


static void
Usage()
{
    fprintf( stderr, "xjsv: Usage: xjsv [-d] [-j js_host]\n" );
}

