/************************************************************************
 *									*
 *	File:  tx_arp.c							*
 *									*
 *	TranlateX (x is silent) ip address to ethernet address using	*
 *	Address Resolution Protocol.					*
 *	The most convenient routine to use is tx_arp itself, but be	*
 *	aware that it can hang (on sleep's) for several minutes waiting	*
 *	for an ARP attempt to time out.  Programs that cannot wait	*
 *	that long or that can't spare SIGALRM should use the routines:	*
 *		tx_arp_init	to start the arp.			*
 *		tx_arp_chk	to check on its progress.		*
 *		tx_arp_wait	to hang waiting for the arp to complete.*
 *				(uses SIGALRM).				*
 *									*
 *	Written 08-Sep-87 by Paul E. McKenney, SRI International.	*
 *									*
 ************************************************************************/
/* @(#)tx_arp.c	1.1 92/12/26 */
/*
 *	$Log:	tx_arp.c,v $
 * Revision 2.0  88/09/23  18:28:23  mckenney
 * .
 * 
 * Revision 1.2  88/09/19  09:15:37  mckenney
 * Added tx_arp_perm and tx_arp_tmp routines
 * 
 */

/* Include files.							*/

#include <stdio.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <net/if.h>
#include <netinet/in.h>
#include <netinet/if_ether.h>
#include <signal.h>

/* Type definitions local to this file.					*/

/* Functions exported from this file.					*/

extern int		tx_arp();		/* TX using ARP.	*/
extern int		tx_arp_chk();		/* TX ARP CHecK.	*/
extern int		tx_arp_init();		/* TX ARP INITiate.	*/
extern int		tx_arp_perm();		/* make entry PERManent.*/
extern int		tx_arp_tmp();		/* make entry TeMPorary.*/
extern int		tx_arp_wait();		/* TX ARP WAIT for done.*/

/* Functions local to this file.					*/

/* Variables exported from this file.					*/

/* Variables local to this file.					*/


/* TranslateX using Address Resolution Protocol.  Internet address	*/
/* (in the form of a sockaddr_in *) is passed in through sa, puts	*/
/* resulting ethernet address into ea.  Returns 0 if all went well,	*/
/* -1 if a system call failed (em is pointed to the name of the system	*/
/* call in this case).							*/
/* Note that no arp will actually be transmitted if the address is	*/
/* already in the cache.						*/

extern int
tx_arp(sa, ea, em)

	struct sockaddr_in *sa;		/* Socket Address.		*/
	struct ether_addr *ea;		/* Ethernet Address.		*/
	char		**em;		/* Error Message pointer.	*/

	{
	int		retval;		/* RETurn VALue.		*/

	/* Initiate the arp.  If this succeeds immediately (because the	*/
	/* arp was already cached), just return.			*/

	if ((retval = tx_arp_init(sa, ea, em)) < 0)
		return (retval);
	if (retval == 1)
		return (0);

	/* Wait for the arp to complete.				*/

	while ((retval = tx_arp_wait(sa, ea, em)) == 0)
		sleep(1);
	return (retval);
	}

/* TranslateX using ARP, CHecK results.  Internet address		*/
/* (in the form of a sockaddr_in *) is passed in through sa, puts	*/
/* resulting ethernet address into ea.  Returns 1 if the desired entry	*/
/* is in the cache, 0 if the arp is in process, and -1 if a system call	*/
/* failed (em is pointed to the name of the system call in this case).	*/

extern int
tx_arp_chk(sa, ea, em)

	struct sockaddr_in *sa;		/* Socket Address.		*/
	struct ether_addr *ea;		/* Ethernet Address.		*/
	char		**em;		/* Error Message pointer.	*/

	{
	static struct arpreq arpreq;	/* ARP REQuest structure.	*/
	int		s;		/* Socket file descriptor.	*/

	/* Set up arp request struct.					*/

	bzero((caddr_t)&arpreq, sizeof(arpreq));
	bcopy(sa, (caddr_t)&arpreq.arp_pa, sizeof(*sa));
        arpreq.arp_pa.sa_family = AF_INET;
        arpreq.arp_ha.sa_family = AF_UNSPEC;

	/* Open a socket to do ioctls on (just to arp!).		*/

        if ((s = socket(AF_INET, SOCK_DGRAM, 0)) == -1)
                {
                *em = "socket";
                return (-1);
                }
 
	/* Do the ioctl, check the result.				*/

	if (ioctl(s, SIOCGARP, (caddr_t)&arpreq) == -1)
		{
		*em = "ioctl SIOCGARP";
		close(s);
		return (-1);
		}

	close(s);
	(void)bcopy(arpreq.arp_ha.sa_data, ea, sizeof(struct ether_addr));
	return ((arpreq.arp_flags & ATF_COM) != 0);
	}

/* TranslateX using ARP, INITiate process.  Internet address		*/
/* (in the form of a sockaddr_in *) is passed in through sa, puts	*/
/* resulting ethernet address into ea.  Returns 1 if the desired entry	*/
/* is already in the cache, 0 if the arp is in process, and -1 if a	*/
/* system call failed (em is pointed to the name of the system call in	*/
/* this case).								*/

int
tx_arp_init(sa, ea, em)

	struct sockaddr_in *sa;		/* Socket Address.		*/
	struct ether_addr *ea;		/* Ethernet Address.		*/
	char		**em;		/* Error Message pointer.	*/

	{
	static struct arpreq arpreq;
	int		discard = -1;
	int		(*oldsig)();
	int		retval;
	int		s = -1;

	/* Set up arp request struct.					*/

	bzero((caddr_t)&arpreq, sizeof(arpreq));
	bcopy(sa, (caddr_t)&arpreq.arp_pa, sizeof(*sa));
        arpreq.arp_pa.sa_family = AF_INET;
        arpreq.arp_ha.sa_family = AF_UNSPEC;

	/* Get a socket suitable for arp'ing.				*/

        if ((s = socket(AF_INET, SOCK_DGRAM, 0)) == -1)
                {
		*em = "socket";
		retval = -1;
		goto cleanup;
                }

	/* Check to see if it is already in cache.			*/

	if ((ioctl(s, SIOCGARP, (caddr_t)&arpreq) != -1) &&
            ((arpreq.arp_flags & ATF_COM) != 0))
		{
		(void)bcopy(arpreq.arp_ha.sa_data, ea,
			    sizeof(struct ether_addr));
		retval = 1;
		goto cleanup;
		}

	/* Send a udp packet to force the system to send an arp.  We	*/
	/* send to the echo port to minimize impact on the net.		*/
	/* Note that we could use an NIT socket for this, but that	*/
	/* would require super-user privileges.				*/

	if ((discard = mk_udp_ncon(em)) < 0)
		{
		retval = discard;
		goto cleanup;
		}
	oldsig = signal(SIGPIPE, SIG_IGN);
	sa->sin_family = AF_INET;
	sa->sin_port = IPPORT_ECHO;
	if (sendto(discard, "abcd", 5, 0, sa, sizeof(*sa)) != 5)
		{
		*em = "sendto";
		retval = -1;
		goto cleanup;
		}
	(void)signal(SIGPIPE, oldsig);

	/* Check cache again just in case . . .				*/

	if (ioctl(s, SIOCGARP, (caddr_t)&arpreq) == -1)
		{
		*em = "ioctl SIOCGARP";
		retval = -1;
		goto cleanup;
		}
        if ((arpreq.arp_flags & ATF_COM) != 0)
		{
		(void)bcopy(arpreq.arp_ha.sa_data, ea,
			    sizeof(struct ether_addr));
		retval = 1;
		goto cleanup;
		}
	retval = 0;

cleanup:

	/* Close file descriptors.					*/

	if (s != -1)
		close(s);
	if (discard != -1)
		close(discard);
	return (retval);
	}

/* TranslateX using ARP, force existing entry to be PERManent.		*/
/* Internet address (in the form of a sockaddr_in *) is passed in	*/
/* through sa, ethernet address (in the form of a struct ether_addr)	*/
/* is optionally passed through ea (if ea is NULL, the address will	*/
/* be looked up in the arp cache).  Returns 1 if the desired entry	*/
/* has been cached, 0 if the arp is in process, and -1 if a system call	*/
/* failed (em is pointed to the name of the system call in this case).	*/

extern int
tx_arp_perm(sa, ea, em)

	struct sockaddr_in *sa;		/* Socket Address.		*/
	struct ether_addr *ea;		/* Ethernet Address.		*/
	char		**em;		/* Error Message pointer.	*/

	{
	static struct arpreq arpreq;	/* ARP REQuest structure.	*/
	int		retval;		/* RETurn VALue.		*/
	int		s;		/* Socket file descriptor.	*/

	/* Set up arp request struct.					*/

	bzero((caddr_t)&arpreq, sizeof(arpreq));
	bcopy(sa, (caddr_t)&arpreq.arp_pa, sizeof(*sa));
        arpreq.arp_pa.sa_family = AF_INET;
        arpreq.arp_ha.sa_family = AF_UNSPEC;

	/* Open a socket to do ioctls on (just to arp!).		*/

        if ((s = socket(AF_INET, SOCK_DGRAM, 0)) == -1)
                {
                *em = "socket";
                return (-1);
                }
 
	/* Do the ioctl, check the result.				*/

	if (((retval = ioctl(s, SIOCGARP, (caddr_t)&arpreq)) == -1) &&
	    (ea == NULL))
		{
		*em = "ioctl SIOCGARP";
		close(s);
		return (-1);
		}
	if ((retval != -1) &&
	    ((arpreq.arp_flags & ATF_COM) == 0))
		{
		close(s);
		return (0);
		}

	/* Set the ethernet address if requested, mark the entry	*/
	/* permanent.							*/

	if (ea != NULL)
		(void)bcopy(ea, (caddr_t)arpreq.arp_ha.sa_data, sizeof(*ea));
	arpreq.arp_flags |= ATF_PERM;
 
	/* Delete the existing entry.					*/

	/* if (ioctl(s, SIOCDARP, (caddr_t)&arpreq) == -1)
		{
		*em = "ioctl SIOCDARP";
		close(s);
		return (-1);
		}
	*/

	/* Do the ioctl, check the result.				*/

	if (ioctl(s, SIOCSARP, (caddr_t)&arpreq) == -1)
		{
		*em = "ioctl SIOCSARP";
		close(s);
		return (-1);
		}
	close(s);
	return (1);
	}

/* TranslateX using ARP, force existing entry to be TeMPorary.		*/
/* Internet address (in the form of a sockaddr_in *) is passed in	*/
/* through sa, ethernet address (in the form of a struct ether_addr)	*/
/* is optionally passed through ea (if ea is NULL, the address will	*/
/* be looked up in the arp cache).  Returns 1 if the desired entry	*/
/* has been cached, 0 if the arp is in process, and -1 if a system call	*/
/* failed (em is pointed to the name of the system call in this case).	*/

extern int
tx_arp_tmp(sa, ea, em)

	struct sockaddr_in *sa;		/* Socket Address.		*/
	struct ether_addr *ea;		/* Ethernet Address.		*/
	char		**em;		/* Error Message pointer.	*/

	{
	static struct arpreq arpreq;	/* ARP REQuest structure.	*/
	int		retval;		/* RETurn VALue.		*/
	int		s;		/* Socket file descriptor.	*/

	/* Set up arp request struct.					*/

	bzero((caddr_t)&arpreq, sizeof(arpreq));
	bcopy(sa, (caddr_t)&arpreq.arp_pa, sizeof(*sa));
        arpreq.arp_pa.sa_family = AF_INET;
        arpreq.arp_ha.sa_family = AF_UNSPEC;

	/* Open a socket to do ioctls on (just to arp!).		*/

        if ((s = socket(AF_INET, SOCK_DGRAM, 0)) == -1)
                {
                *em = "socket";
                return (-1);
                }
 
	/* Do the ioctl, check the result.				*/

	if (((retval = ioctl(s, SIOCGARP, (caddr_t)&arpreq)) == -1) &&
	    (ea == NULL))
		{
		*em = "ioctl SIOCGARP";
		close(s);
		return (-1);
		}
	if ((retval != -1) &&
	    ((arpreq.arp_flags & ATF_COM) == 0))
		{
		close(s);
		return (0);
		}

	/* Set the ethernet address if requested, mark the entry	*/
	/* permanent.							*/

	if (ea != NULL)
		(void)bcopy(ea, (caddr_t)arpreq.arp_ha.sa_data, sizeof(*ea));
	arpreq.arp_flags &= ~ATF_PERM;
 
	/* Delete the existing entry.					*/

	/* if (ioctl(s, SIOCDARP, (caddr_t)&arpreq) == -1)
		{
		*em = "ioctl SIOCDARP";
		close(s);
		return (-1);
		}
	*/

	/* Do the ioctl, check the result.				*/

	if (ioctl(s, SIOCSARP, (caddr_t)&arpreq) == -1)
		{
		*em = "ioctl SIOCSARP";
		close(s);
		return (-1);
		}
	close(s);
	return (1);
	}

/* TranslateX using ARP, WAIT for completion.  Internet address		*/
/* (in the form of a sockaddr_in *) is passed in through sa, puts	*/
/* resulting ethernet address into ea.  Returns 1 all went well, and -1	*/
/* if a system call failed (em is pointed to the name of the system	*/
/* call in this case).  Note that this routine can hang for several	*/
/* minutes while the system waits for an arp response, and that this	*/
/* routine (indirectly) uses SIGALRM.					*/

int
tx_arp_wait(sa, ea, em)

	struct sockaddr_in *sa;		/* Socket Address.		*/
	struct ether_addr *ea;		/* Ethernet Address.		*/
	char		**em;		/* Error Message pointer.	*/

	{
	int		retval;

	/* check status until we get a definite answer.			*/

	while ((retval = tx_arp_chk(sa, ea, em)) == 0)
		sleep(1);
	return (retval);
	}
