/*
 * ripwatch - RIP packet monitoring program.
 *
 */

#include <sys/param.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <net/if.h>
#include <netinet/in.h>
#include <signal.h>
#include <errno.h>
#include <stdio.h>

#ifdef sun
#include <net/nit_if.h>
#include <net/nit_buf.h>

#define USE_NIT

char   *devices[] = {
    "le0", "ie0", "le1", "ie1", "le2", "ie2",
    "le3", "ie3", "le4", "ie4", 0
    };

#endif /* sun */

#ifdef ultrix
#include <net/pfilt.h>

#define USE_PFILT

char   *devices[] = {
    "pf0", "pf1", "pf2", "pf3", "pf4", "pf5",
    "pf6", "pf7", "pf8", "pf9", 0
    };

#endif /* ultrix */

#include "ripwatch.h"

char *	pname;					/* program name		 */

FILE *	logfp;					/* log file pointer	 */

Counter pkt_total = 0;				/* total packets seen	 */
Counter pkt_drops = 0;				/* total packets dropped */
Counter int_pkt_total = 0;			/* packets this interval */
Counter int_pkt_drops = 0;			/* dropped this interval */
Counter dst_pkt_total = 0;			/* total pkts to host	 */
Counter int_dst_pkt_total = 0;			/* pkts to host this int */

int     if_fd[MAXINTERFACES];			/* nit device file desc	 */

int     srcflag = 0;				/* "-src" specified	 */
int     dstflag = 0;				/* "-dst" specified	 */
int     allflag = 0;				/* "-all" specified	 */
int     allintf = 1;				/* "-allif" specified	 */
int	debug	= 0;				/* "-d" specified	 */
int     logging = 0;				/* 1 when logging on	 */
int     learnfs = 0;				/* learn other servers	 */
int     runtime = RUNTIME;			/* run time		 */
int     truncation = 200;			/* pkt trunc len - magic */

struct timeval starttime;			/* time we started	 */
int     now;					/* current seconds	 */

int     ninterfaces;				/* number of interfaces	 */
int     non_prom = 0;				/* 0 = promiscious	 */

int	num_net_intf;
struct net_intf net_intf[MAXINTERFACES];

u_long  thisdst = 0;				/* cached IP dst of pkt	 */

char *	dumpfile = DUMPFILE;			/* dump/load file name	 */
char *	logfile = LOGFILE;			/* log file name	 */

PacketCounter pkt_counters[PKT_NCOUNTERS];	/* packet counters	 */

#ifdef ultrix
void
fpe_warn()
{
    extern void finish();

    fprintf( stderr, "em_ripwatch: mystery bug encountered.\n" );
    finish(-1);
}
#endif	/* ultrix */


main( argc, argv )
int     argc;
char  **argv;
{
    int     i;
    struct sigvec sv;
    char   *device = NULL;
    extern void finish();
    extern void ripwatch();

    setlinebuf( stdout );
    printf( " em_ripwatch: started\n" );

    pname = *argv;

    /* Process arguments. */

    while (--argc)
	{
	if (**++argv != '-')
	    usage();

	/* Device to use. */

	if (!strcmp(*argv, "-dev"))
	    {
	    if (--argc <= 0)
		usage();

	    device = *++argv;
	    continue;
	    }

	/* Turn off promiscious mode on NIT */

	if (!strcmp(*argv, "-p"))
	    {
	    non_prom++;
	    continue;
	    }

	/* Log file name. */

	if (!strcmp(*argv, "-lf"))
	    {
	    if (--argc <= 0)
		usage();

	    logfile = *++argv;
	    continue;
	    }

	/* Change run time. */

	if (!strcmp(*argv, "-t"))
	    {
	    if (--argc <= 0)
		usage();

	    runtime = atoi(*++argv);
	    continue;
	    }

	/* Turn on logging. */

	if (!strcmp(*argv, "-l"))
	    {
	    logging++;
	    continue;
	    }

	/* Watch all traffic. */

	if (!strcmp(*argv, "-all"))
	    {
	    allflag++;
	    continue;
	    }

	/* Use all interfaces. */

	if (!strcmp(*argv, "-allif"))
	    {
	    allintf++;
	    continue;
	    }

	usage();
	}

    /* Trap signals so we can clean up. */

    sv.sv_handler = finish;
    sv.sv_mask = sv.sv_flags = 0;

    (void) sigvec(SIGINT, &sv, (struct sigvec *) 0);
    (void) sigvec(SIGQUIT, &sv, (struct sigvec *) 0);

#ifdef ultrix
    sv.sv_handler = fpe_warn;
    (void) sigvec(SIGFPE, &sv, (struct sigvec *) 0);
#endif /* ultrix */

#ifdef USE_NIT
    /* Set up the network interface tap right away, since we probably need
       super-user permission. */

    if (allintf)
	{
	ninterfaces = 0;
	for (i = 0; devices[i] != NULL; i++)
	    {
	    if_fd[ninterfaces] = setup_nit_dev(&devices[i]);

	    if (if_fd[ninterfaces] >= 0)
		ninterfaces++;
	    }
	}
    else
	{
	if_fd[0] = setup_nit_dev(&device);

	if (if_fd[0] < 0)
	    {
	    perror(device);
	    finish(-1);
	    }

	ninterfaces = 1;
	}
#endif /* USE_NIT */

#ifdef USE_PFILT
    /* Set up the packet filter interface now, although we don't need
       super-user permission. */

    if (allintf)
	{
	ninterfaces = 0;
	for (i = 0; devices[i] != NULL; i++)
	    {
	    if_fd[ninterfaces] = setup_pfilt_dev(&devices[i]);

	    if (if_fd[ninterfaces] >= 0)
		ninterfaces++;
	    }
	}
    else
	{
	if_fd[0] = setup_pfilt_dev(&device);

	if (if_fd[0] < 0)
	    {
	    perror(device);
	    finish(-1);
	    }

	ninterfaces = 1;
	}
#endif /* USE_PFILT */

    if (ninterfaces < 1)
	{
	fprintf(stderr, "em_ripwatch: %s: no valid interfaces.\n", pname);
	finish(-1);
	}

    /* Now lose super-user permission, since we don't need it for anything
       else. */

    (void) setreuid(getuid(), getuid());

    GetIntf();

    if ( debug )
	{
	if (allintf)
	    (void) printf( " em_ripwatch: all interfaces;\n" );
	else
	    (void) printf( " em_ripwatch: interface %s;\n", device );

	printf( " em_ripwatch: log to \"%s\" (logging %s);\n", logfile,
		    (logging ? "on" : "off" ));
	printf( " em_ripwatch: dump/load to/from \"%s\";\n", dumpfile );
	}

    /* Load previous data from dumpfile */
/*
 *  printf( " em_ripwatch: loading previous data from \"%s\";\n", dumpfile );
 *  loadtable();
 *  printf( " em_ripwatch: data load complete\n" );
 */

    /* Set up the packet counters. */

    setup_pkt_counters();

    /* Open log file if logging is on. */

    if (logging)
	{
	if ((logfp = fopen(logfile, "a")) == NULL)
	    {
	    error(logfile);
	    finish(-1);
	    }

	(void) fprintf( logfp, "#\n# startlog\n#\n" );
	(void) fprintf( logfp, "# ripwatch log file\n" );
	}

    /* Go watch packets.  Never returns. */

    ripwatch();
    return (0);
}


/*
 * ripwatch - main packet reading loop.
 */

void
ripwatch()
{
    int     cc;
    char *  buf;
    int     i;
    char *  malloc();
    fd_set  readfds;
    struct sigvec sv;
    struct timeval tv;
    extern void finish();
    struct itimerval itv;
    register char * bp;
    register char * cp;
    register char * bufstop;

#ifdef USE_NIT
    int     tdrops[MAXINTERFACES];
    struct nit_bufhdr *hdrp;
    struct nit_ifdrops *ndp;

#endif /* USE_NIT */

#ifdef USE_PFILT
    struct enstamp stamp;
    int     datalen;

#endif /* USE_PFILT */

    /* Allocate a buffer so it's properly aligned for casting to structure
       types. */

    if ((buf = malloc(NIT_CHUNKSIZE)) == NULL)
	{
	(void) fprintf( stderr, "em_ripwatch: %s: out of memory.\n", pname );
	finish(-1);
	}

    /* Set up the alarm handler. */

    sv.sv_handler = finish;
    sv.sv_mask = sv.sv_flags = 0;

    (void) sigvec(SIGALRM, &sv, (struct sigvec *) 0);

    /* Set up the alarm clock. */

/*  (void) bzero((char *) &itv, sizeof(struct itimerval)); */
    (void) bzero(&itv, sizeof(struct itimerval));

    itv.it_interval.tv_sec = runtime;
    itv.it_interval.tv_usec = 0;

    itv.it_value = itv.it_interval;

    (void) setitimer(ITIMER_REAL, &itv, (struct itimerval *) 0);

    /* Set the start time. */

    (void) gettimeofday(&starttime, (struct timezone *) 0);
    now = starttime.tv_sec;

#ifdef USE_NIT
    /* Flush the read queue of any packets that accumulated during setup
       time. */

    for (i = 0; i < ninterfaces; i++)
	{
	flush_nit(if_fd[i]);
	tdrops[i] = 0;
	}

    for (;;)
	{
	FD_ZERO(&readfds);

	for (i = 0; i < ninterfaces; i++)
	    FD_SET(if_fd[i], &readfds);

	/* See which nets have packets to read. */

	cc = select(NFDBITS, &readfds, (fd_set *) 0, (fd_set *) 0,
		    (struct timeval *) 0);
	if ((cc < 0) && (errno != EINTR))
	    {
	    error("select");
	    finish(-1);
	    }
	if (cc == 0)
	    continue;

	/* For each interface... */

	for (i = 0; i < ninterfaces; i++)
	    {
	    /* Nothing to read. */

	    if (!FD_ISSET(if_fd[i], &readfds))
		continue;

	    /* Now read packets from the nit device. */

	    if ((cc = read(if_fd[i], buf, NIT_CHUNKSIZE)) <= 0)
		continue;

	    bufstop = buf + cc;
	    bp = buf;

	    /* Loop through the chunk, extracting packets. */

	    while (bp < bufstop)
		{
		cp = bp;

		/* Get the nit header. */
		hdrp = (struct nit_bufhdr *) cp;
		cp += sizeof(struct nit_bufhdr);

		/* Get the number of dropped packets. */
		ndp = (struct nit_ifdrops *) cp;
		cp += sizeof(struct nit_ifdrops);

		int_pkt_drops += ndp->nh_drops - tdrops[i];
		pkt_drops += ndp->nh_drops - tdrops[i];
		tdrops[i] = ndp->nh_drops;

		/* Filter the packet. */

		pkt_filter(cp, hdrp->nhb_msglen);

		/* Skip over this packet. */

		bp += hdrp->nhb_totlen;
		}
	    }
	}
#endif /* USE_NIT */

#ifdef USE_PFILT
    /* Flush the read queue of any packets that accumulated during setup
       time. */

    for (i = 0; i < ninterfaces; i++)
	flush_pfilt(if_fd[i]);

    for (;;)
	{
	FD_ZERO(&readfds);

	for (i = 0; i < ninterfaces; i++)
	    FD_SET(if_fd[i], &readfds);

	/* See which interfaces have any packets to read. */

	cc = select(NFDBITS, &readfds, (fd_set *) 0, (fd_set *) 0, 0);

	if ((cc < 0) && (errno != EINTR))
	    {
	    error("select");
	    finish(-1);
	    }
	if (cc == 0)
	    continue;

	/* Now read packets from the packet filter device. */

	for (i = 0; i < ninterfaces; i++)
	    {
	    if (!FD_ISSET(if_fd[i], &readfds))
		continue;

	    if ((cc = read(if_fd[i], buf, NIT_CHUNKSIZE)) < 0)
		{
		lseek(if_fd[i], 0L, 0);

		/* Might have read MAXINT bytes.  Try again. */

		if ((cc = read(if_fd[i], buf, NIT_CHUNKSIZE)) < 0)
		    {
		    error("pfilt read");
		    finish(-1);
		    }
		}

	    bp = buf;

	    /* Loop through buffer, extracting packets. */

	    while (cc > 0)
		{
		/* Avoid alignment issues. */

		(void) bcopy(bp, &stamp, sizeof(stamp));

		/* Treat entire buffer as garbage. */

		if (stamp.ens_stamplen != sizeof(stamp))
		    break;

		/* Get the number of dropped packets. */

		int_pkt_drops += stamp.ens_dropped;
		pkt_drops += stamp.ens_dropped;

		/* Filter the packet. */

		datalen = stamp.ens_count;

		if (datalen > truncation)
		    datalen = truncation;

		pkt_filter(&(bp[sizeof(stamp)]), datalen);

		/* Skip over this packet. */

		if (cc == (datalen + sizeof(stamp)))
		    break;

		datalen = ENALIGN(datalen);
		datalen += sizeof(stamp);
		cc -= datalen;
		bp += datalen;
		}
	    }
	}
#endif /* USE_PFILT */

}

